/*
 * cipher - a more secure replacement for UNIX crypt program 
 *
 * Generic block encipherment/decipherment program 
 * modified specifically for the Data Encryption Standard using
 * Cipher Block Feedback (CFB) mode as described in FIPS pub 81.
 *
 * See also FIPS PUB 74.
 *
 * Permission is given copy and use this program provided it is not sold
 * for profit.
 *
 * Written by David A. Barrett (barrett%asgard@boulder.Colorado.EDU)
 */

#define TRUE 1
#define FALSE 0

#include <unistd.h>	/* POSIX: read write */
#include <stdlib.h>	/* AES: getopt ANSII C: exit */
#include <stdio.h>
#include <string.h>
#include "cipher.h"

extern int optind;
extern char *optarg;


#define	IO_BUFSIZE	1024		/* small for data cache speed */
#define KEYBUFLEN	129		/* allow long keys */

typedef struct {
  char	   iv[DES_BLOCKSIZE];
  keyType key;
  int	   ivlen;
  int	   decr;
} KeyItem;

KeyItem key_item ;
char    bufs[2][IO_BUFSIZE];

typedef unsigned char block[DES_BLOCKSIZE];

/*
 * Keys which could accidently reproduce the plaintext with multiple-encryption
 */
block badKeys[16] = {
  { 0xe0, 0x00, 0xe0, 0x00, 0xf0, 0x00, 0xf0, 0x00 },
  { 0x00, 0xe0, 0x00, 0xe0, 0x00, 0xf0, 0x00, 0xf0 }, 
  { 0xfe, 0x1e, 0xfe, 0x1e, 0xfe, 0x0e, 0xfe, 0x0e },
  { 0x1e, 0xfe, 0x1e, 0xfe, 0x0e, 0xfe, 0x0e, 0xfe }, 
  { 0xe0, 0x1e, 0xe0, 0x1e, 0xf0, 0x0e, 0xf0, 0x0e },
  { 0x1e, 0xe0, 0x1e, 0xe0, 0x0e, 0xf0, 0x0e, 0xf0 }, 
  { 0x00, 0xfe, 0x00, 0xfe, 0x00, 0xfe, 0x00, 0xfe },
  { 0xfe, 0x00, 0xfe, 0x00, 0xfe, 0x00, 0xfe, 0x00 },
  { 0x00, 0x1e, 0x00, 0x1e, 0x00, 0x0e, 0x00, 0x0e },
  { 0x1e, 0x00, 0x1e, 0x00, 0x0e, 0x00, 0x0e, 0x00 }, 
  { 0xe0, 0xfe, 0xe0, 0xfe, 0xf0, 0xfe, 0xf0, 0xfe },
  { 0xfe, 0xe0, 0xfe, 0xe0, 0xfe, 0xf0, 0xfe, 0xf0 }, 
  { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },	/* self-dual */
  { 0xfe, 0xfe, 0xfe, 0xfe, 0xfe, 0xfe, 0xfe, 0xfe },	/* self-dual */
  { 0xe0, 0xe0, 0xe0, 0xe0, 0xf0, 0xf0, 0xf0, 0xf0 },	/* self-dual */
  { 0x1e, 0x1e, 0x1e, 0x1e, 0x0e, 0x0e, 0x0e, 0x0e },	/* self-dual */
};

/*
 * Return nonzero if the key is one of the well-known "weak" keys (FIPS Pub 74)
 */
int weakKey(keybits)
     char *keybits;
{
  register block *keyp = &(badKeys[0]);	/* common compiler bug here */
  register unsigned count = sizeof(badKeys) / sizeof(badKeys[0]);
  register unsigned i;

  do {
    for (i = 0; i < DES_BLOCKSIZE; i++) {
      if ((((*keyp)[i] ^ keybits[i]) & 0xfe) != 0) goto notfound;
    }
    break;
  notfound:
    keyp++;
  } while (--count != 0);
  return count;
}



/*
 * Because we are using CFB, we always encrypt.  Since we'd like to be
 * able to run the chips in either mode, perhaps CBC would be better.
 * If CBC is used though, the last block must be handled in OFB mode.  Sigh.
 *
 * Returns:  
 *  -1 - out of memory (program exits, and error reported)
 *   0 - success
 *   1 - key is a weak key
 */
int setKey(kp, keybits, keystr, mode)
     register keyType *kp;
     char *keybits, *keystr;
     int  mode;
{
  register int res;

  if (keystr != (char *) 0) {
    desKey(keybits, keystr, ' ');		/* space padded */
    memset(keystr, '\0', strlen(keystr));	/* blank out key asap */
  }

  res = weakKey(keybits);

  makeKey(kp, keybits, DES_BLOCKSIZE, mode) ;	/* 0 means encrypt mode */
  memset(keybits, '\0', DES_BLOCKSIZE) ;  	/* zero out bits asap */

  if (*kp == (keyType) 0) {
    fprintf (stderr, "Couldn't allocate memory for encryption key\n") ;
    exit(1);
  }
  return res;
}






int main (argc, argv)
     int argc;
     char *argv[];
{
  int i, ch, rcount, wcount = -1; 
  int keystrlen = 0, keylen = 0;
  char *chp, *ibuf = bufs[1], *obuf = bufs[0];
  char keybits[DES_BLOCKSIZE], ivbits[DES_BLOCKSIZE];
  char keybuf[KEYBUFLEN], ivbuf[DES_BLOCKSIZE];

  while ((ch = getopt(argc, argv, "devni:k:x:")) != EOF)
    switch (ch) {
    case 'd':  
      key_item.decr = TRUE ;		/* decrypt */
      break;
    case 'e':  
      key_item.decr = FALSE ;		/* encrypt */
      break;
    default:
      fprintf (stderr, "usage: {-de}\n");
      exit (0) ;
      break ;
    }
  argc -= optind;
  argv += optind;

  key_item.ivlen = DES_BLOCKSIZE;
  memset (key_item.iv, '\0', DES_BLOCKSIZE);

  keystrlen = getkey ("Input Key:", keybuf, KEYBUFLEN-1);
  if (keystrlen <= 0) {
    fprintf (stderr, "Couldn't get key\n");
    exit (-1) ;
  }

  keybuf[keystrlen] = '\0';
  i = setKey (&key_item.key, keybits, keybuf, 0);
  if (i) {
    fprintf (stderr, "Weak key generated; use another one\n") ;
    exit (-1) ;
  }

  do {
    rcount = read(0, obuf, IO_BUFSIZE);
    if (rcount <= 0) break;

    chp = obuf;
    obuf = ibuf;
    ibuf = chp;
    key_item.ivlen =
      desCFB (obuf, ibuf, rcount, key_item.iv, key_item.ivlen,
	      key_item.key, key_item.decr) ;
    wcount = write (1, obuf, rcount) ;
  } while (wcount == rcount) ;

  desDestroyKey (&key_item.key) ;
  exit (0) ;
}
